/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Copyright (C) 2020-2022 Loongson Technology Corporation Limited
 */

#ifndef __LOONGARCH_KVM_CSR_H__
#define __LOONGARCH_KVM_CSR_H__
#include <asm/loongarchregs.h>
#include <asm/kvm_host.h>
#include <asm/kvmcpu.h>
#include <asm/watch.h>
#include <linux/uaccess.h>
#include <linux/kvm_host.h>

#define kvm_read_hw_gcsr(id)			gcsr_read(id)
#define kvm_write_hw_gcsr(csr, id, val)		gcsr_write(val, id)

int lvz_getcsr(struct kvm_vcpu *vcpu, unsigned int id, u64 *v, int force);
int lvz_setcsr(struct kvm_vcpu *vcpu, unsigned int id, u64 *v, int force);
unsigned long lvz_gpsi_read_csr(struct kvm_vcpu *vcpu, int csrid);
void lvz_gpsi_write_csr(struct kvm_vcpu *vcpu, int csrid,
	unsigned long val);
void lvz_gpsi_change_csr(struct kvm_vcpu *vcpu, int csrid,
	unsigned long csr_mask, unsigned long val);
int lvz_gpsi_iocsr(larch_inst inst,
		struct kvm_run *run, struct kvm_vcpu *vcpu);

static inline void kvm_save_hw_gcsr(struct loongarch_csrs *csr, int gid)
{
	csr->csrs[gid] = gcsr_read(gid);
}

static inline void kvm_restore_hw_gcsr(struct loongarch_csrs *csr, int gid)
{
	gcsr_write(csr->csrs[gid], gid);
}

static inline unsigned long kvm_read_sw_gcsr(struct loongarch_csrs *csr, int gid)
{
	return csr->csrs[gid];
}

static inline void kvm_write_sw_gcsr(struct loongarch_csrs *csr, int gid, unsigned long val)
{
	csr->csrs[gid] = val;
}

static inline void kvm_set_sw_gcsr(struct loongarch_csrs *csr, int gid, unsigned long val)
{
	csr->csrs[gid] |= val;
}

static inline void kvm_change_sw_gcsr(struct loongarch_csrs *csr, int gid, unsigned mask,
	unsigned long val)
{
	unsigned long _mask = mask;
	csr->csrs[gid] &= ~_mask;
	csr->csrs[gid] |= val & _mask;
}


#define GET_HW_GCSR(id, csrid, v)				\
	do {							\
		if (csrid == id) {				\
			*v = (long)kvm_read_hw_gcsr(csrid);	\
			return 0;				\
		}						\
	} while (0)

#define GET_SW_GCSR(csr, regid, csrid, v)			\
	do {							\
		if (csrid == id) {				\
			*v = kvm_read_sw_gcsr(csr, csrid);	\
			return 0;				\
		}						\
	} while (0)

#define SET_HW_GCSR(csr, id, csrid, v)				\
	do {							\
		if (csrid == id) {				\
			kvm_write_hw_gcsr(csr, csrid, *v);	\
			return 0;				\
		}						\
	} while (0)

#define SET_SW_GCSR(csr, id, csrid, v)				\
	do {							\
		if (csrid == id) {				\
			kvm_write_sw_gcsr(csr, csrid, *v);	\
			return 0;				\
		}						\
	} while (0)
int lvz_iocsr_init(struct kvm *kvm);
struct kvm_iocsr_entry *kvm_find_iocsr(struct kvm *kvm, u32 addr);

int lvz_iocsr_set(struct kvm *kvm, struct kvm_iocsr_entry *__user argp);
int lvz_iocsr_get(struct kvm *kvm, struct kvm_iocsr_entry *__user argp);

#define KVM_PMU_PLV_ENABLE      (CSR_PERFCTRL_PLV0 |            \
					CSR_PERFCTRL_PLV1 |     \
					CSR_PERFCTRL_PLV2 |     \
					CSR_PERFCTRL_PLV3)

#define CASE_GPSI_WRITE_HW_PMU(vcpu, csr, id, csrid, v)                                 \
	do {                                                                            \
		if (csrid == id) {                                                      \
			if (v & KVM_PMU_PLV_ENABLE) {                                   \
				write_csr_gcfg(read_csr_gcfg() | CSR_GCFG_GPERF);       \
				kvm_write_hw_gcsr(csr, csrid, v | CSR_PERFCTRL_GMOD);   \
				vcpu->arch.aux_inuse |= KVM_LARCH_PERF;                 \
				return ;                                                \
			} else {                                                        \
				kvm_write_sw_gcsr(csr, csrid, v);                       \
				return;                                                 \
			}                                                               \
		}                                                                       \
	} while (0)

/* config fwpc/mwpc debug csr for guest */
#define kvm_enable_hw_breakpoint(hw_breakpoint_csr)     \
	do {                                            \
		int csr_val;                            \
		csr_val = csr_read64(hw_breakpoint_csr); \
		csr_val |= (csr_val & 0x3f) << 16;      \
		csr_write64(csr_val, hw_breakpoint_csr); \
	} while (0)

#define CASE_GPSI_WRITE_DATA_BREAKPOINT(vcpu, csr, regid, csrid, val)   \
	do {                                                            \
		if (regid == csrid) {                                   \
			kvm_enable_hw_breakpoint(LOONGARCH_CSR_MWPC);   \
			kvm_write_hw_gcsr(csr, csrid, val);             \
			vcpu->arch.aux_inuse |= KVM_LARCH_DATA_HWBP;    \
			return ;                                        \
		}                                                       \
	} while (0)

#define CASE_GPSI_WRITE_INST_BREAKPOINT(vcpu, csr, regid, csrid, val)   \
	do {                                                            \
		if (regid == csrid) {                                   \
			kvm_enable_hw_breakpoint(LOONGARCH_CSR_FWPC);   \
			kvm_write_hw_gcsr(csr, csrid, val);             \
			vcpu->arch.aux_inuse |= KVM_LARCH_INST_HWBP;    \
			return ;                                        \
		}                                                       \
	} while (0)

static inline void disable_data_hwbp(void)
{
	int clear_mwpc_mask = ~0xF0000;
	csr_write64(csr_read64(LOONGARCH_CSR_MWPC) & clear_mwpc_mask, LOONGARCH_CSR_MWPC);
}

static inline void disable_inst_hwbp(void)
{
	int clear_fwpc_mask = ~0xF0000;
	csr_write64(csr_read64(LOONGARCH_CSR_FWPC) & clear_fwpc_mask, LOONGARCH_CSR_FWPC);
}

static inline void enable_data_hwbp(void)
{
	kvm_enable_hw_breakpoint(LOONGARCH_CSR_MWPC);
}

static inline void enable_inst_hwbp(void)
{
	kvm_enable_hw_breakpoint(LOONGARCH_CSR_FWPC);
}

static inline void clear_current_thread_hwbp(struct kvm_vcpu *vcpu)
{
	if (vcpu->guest_debug & KVM_GUESTDBG_USE_HW_BP || vcpu->arch.aux_inuse & KVM_LARCH_HWBP) {
		if (test_bit(TIF_LOAD_WATCH, &task_thread_info(current)->flags)) {
			loongarch_clear_watch_registers();
		}
	}
}

static inline void clear_hypervisor_guest_hwbp(struct kvm_vcpu *vcpu)
{
	if (vcpu->guest_debug & KVM_GUESTDBG_USE_HW_BP || vcpu->arch.aux_inuse & KVM_LARCH_HWBP) {
		loongarch_clear_watch_registers();
		/* Install hardware breakpoint csr if needed */
		if (test_bit(TIF_LOAD_WATCH, &task_thread_info(current)->flags)) {
			loongarch_install_watch_registers(current);
		} else if (vcpu->guest_debug & KVM_GUESTDBG_USE_HW_BP) {
			csr_xchg64(0, CSR_PRMD_PWE, LOONGARCH_CSR_PRMD);
		}
	}
}

static inline void kvm_enable_hypervisor_hwbp(struct kvm_vcpu *vcpu)
{
	if (vcpu->guest_debug & KVM_GUESTDBG_USE_HW_BP) {
		csr_xchg64(CSR_PRMD_PWE, CSR_PRMD_PWE, LOONGARCH_CSR_PRMD);
	}
}

static inline void restore_hypervisor_hw_breakpoint(struct kvm_vcpu *vcpu)
{
	int i;

	/* enable hw breakpoint */
	if (vcpu->guest_debug & KVM_GUESTDBG_USE_HW_BP) {
		for (i = 0; i < vcpu->arch.guest_debug.inst_bp_nums; i++) {
			watch_csrwr(vcpu->arch.guest_debug.inst_breakpoint[i].addr, LOONGARCH_CSR_IB0ADDR + 8 * i);
			watch_csrwr(vcpu->arch.guest_debug.inst_breakpoint[i].mask, LOONGARCH_CSR_IB0MASK + 8 * i);
			watch_csrwr(vcpu->arch.guest_debug.inst_breakpoint[i].asid, LOONGARCH_CSR_IB0ASID + 8 * i);
			watch_csrwr(vcpu->arch.guest_debug.inst_breakpoint[i].ctrl, LOONGARCH_CSR_IB0CTL + 8 * i);
		}
		for (i = 0; i < vcpu->arch.guest_debug.data_bp_nums; i++) {
			watch_csrwr(vcpu->arch.guest_debug.data_breakpoint[i].addr, LOONGARCH_CSR_DB0ADDR + 8 * i);
			watch_csrwr(vcpu->arch.guest_debug.data_breakpoint[i].mask, LOONGARCH_CSR_DB0MASK + 8 * i);
			watch_csrwr(vcpu->arch.guest_debug.data_breakpoint[i].asid, LOONGARCH_CSR_DB0ASID + 8 * i);
			watch_csrwr(vcpu->arch.guest_debug.data_breakpoint[i].ctrl, LOONGARCH_CSR_DB0CTL + 8 * i);
		}
	}
}

#endif	/* __LOONGARCH_KVM_CSR_H__ */
